/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * Copyright (C) 2006-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://openbluesky.kenai.com
 * SCM: https://bitbucket.org/tidalwave/openbluesky-src
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.nodes;

import javax.annotation.Nonnull;
import org.openide.util.Lookup;
import org.openide.nodes.Node;
import it.tidalwave.util.Finder;
import it.tidalwave.role.Composite;
import it.tidalwave.netbeans.nodes.role.impl.DecoratorChildren;
import it.tidalwave.netbeans.nodes.role.impl.RearrangingChildren;

/*********************************************************************************************************************
 *
 * A specialization of {@link GenericFilterNode} which provides its own implementation of 
 * {@link FilterNode.Children}. Thus, to create a hierarchy of filter nodes, you just need to wrap the root.
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class GenericDecoratorNode extends GenericFilterNode
  {
    /*******************************************************************************************************************
     *
     * Creates a {@code GenericDecoratorNode} given the delegate node.
     * 
     * @param  delegate     the delegate node
     *
     ******************************************************************************************************************/
    public GenericDecoratorNode (final @Nonnull Node delegate)
      {
        this(delegate, new Object[0]);
      }

    /*******************************************************************************************************************
     *
     * Creates a {@code GenericDecoratorNode} given the delegate node, and (optional) extra arguments
     * (meaningful if your subclasses need them).  The extra arguments are passed as they are to constructor of children 
     * nodes.
     * 
     * @param  delegate     the delegate node
     * @param  extraArgs    the optional extra arguments
     *
     ******************************************************************************************************************/
    public GenericDecoratorNode (final @Nonnull Node delegate, 
                                 final @Nonnull Object ... extraArgs)
      {
        super(delegate, new DecoratorChildren(delegate, extraArgs));
      }

    /*******************************************************************************************************************
     *
     * Creates a {@code GenericDecoratorNode} given the delegate node, a {@link Lookup} and (optional) extra arguments
     * (meaningful if your subclasses need them).  The extra arguments are passed as they are to constructor of children nodes.
     * 
     * @param  delegate     the delegate node
     * @param  lookup       the {@code Lookup}
     * @param  extraArgs    the optional extra arguments
     *
     ******************************************************************************************************************/
    protected GenericDecoratorNode (final @Nonnull Node delegate,
                                    final @Nonnull Lookup lookup,
                                    final @Nonnull Object ... extraArgs)
      {
        super(delegate, createChildren(delegate, lookup, extraArgs), lookup);
      }

    @Nonnull
    private static org.openide.nodes.Children createChildren (final @Nonnull Node delegateNode, 
                                                              final @Nonnull Lookup lookup,
                                                              final @Nonnull Object[] extraArgs) 
      {
        //
        // If we have a different Composite than the delegate, we have to compute children on our own.
        // Otherwise, we use a DecoratorChildren as usual per FilterNodes that do not manipulate children.
        //
        final Composite<?, Finder<?>> newComposite = lookup.lookup(Composite.class);
        final Composite<?, Finder<?>> oldComposite = delegateNode.getLookup().lookup(Composite.class);
        
        if (newComposite == oldComposite)
          {
            return new DecoratorChildren(delegateNode, extraArgs);
          }
        else
          {
            // Can't occur, since newComposite != oldComposite, and removing a Composite is not supported
            assert newComposite != null : "newComposite is null";
            // TODO: should reuse the existing nodes. Should use the DecoratorChildren as usual, but 
            // rearranging nodes according to the order provided by the Composite, eventually ignoring things
            // not created by the composite. New items created in the Composite should be mapped to GenericNode.
                          
            return new RearrangingChildren(delegateNode, newComposite, extraArgs);
          }
      }
  }

