/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * Copyright (C) 2006-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://openbluesky.kenai.com
 * SCM: https://bitbucket.org/tidalwave/openbluesky-src
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.nodes;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.beans.BeanInfo;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.awt.Image;
import javax.swing.Action;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.nodes.FilterNode;
import org.openide.nodes.Node;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.role.Displayable;
import it.tidalwave.role.IconProvider;
import it.tidalwave.role.MutableIconProvider;
import it.tidalwave.netbeans.nodes.role.ActionProvider;
import it.tidalwave.netbeans.nodes.role.Describable;
import static it.tidalwave.role.Displayable.Displayable;
import static it.tidalwave.netbeans.nodes.role.ActionProvider.ActionProvider;
import static it.tidalwave.netbeans.nodes.role.Describable.Describable;
import static it.tidalwave.netbeans.nodes.role.HtmlDisplayName.HtmlDisplayName;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class GenericFilterNode extends FilterNode
  {
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public GenericFilterNode (final @Nonnull Node original)
      {
        super(original);
        installIconChangeListener();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public GenericFilterNode (final @Nonnull Node original, final @Nonnull org.openide.nodes.Children children)
      {
        super(original, children);
        installIconChangeListener();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public GenericFilterNode (final @Nonnull Node original,
                              final @Nonnull org.openide.nodes.Children children,
                              final @Nonnull Lookup lookup)
      {
        super(original, children, lookup);
        installIconChangeListener();
      }

    // Overrides some behaviours that depend on the Lookup, that could be different than the one of the original Node
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String getDisplayName() 
      {
        return getName(super.getDisplayName());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String getHtmlDisplayName() 
      {
        return getLookup().lookup(HtmlDisplayName) != null ? getDisplayName() : null;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String getName() 
      {
        return getName(super.getName());
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String getShortDescription() 
      {
        final Node delegate = getOriginal();
        //
        // If the delegate is not a GenericNode, it is not using a Displayable for its icon, thus it is directly
        // providing its own icon. It could even have an Displayable in its Lookup as it could be a FilterNode; for
        // this reason there's this explicit check.
        //
        if (!(delegate instanceof NodePresentationModel) && !(delegate instanceof GenericFilterNode))
          {
            return super.getShortDescription();
          }

        final Describable describable = getLookup().lookup(Describable);
        return (describable != null) ? describable.getDescription() : super.getShortDescription();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Image getIcon (final int size)
      {
        return getIcon(size, super.getIcon(size));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Image getOpenedIcon (final int type)
      {
        return getIcon(type, super.getIcon(type));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    private String getName (final @Nonnull String fallBackName)
      {
        final Node delegate = getOriginal();
        //
        // If the delegate is not a GenericNode, it is not using a Displayable for its icon, thus it is directly
        // providing its own icon. It could even have an Displayable in its Lookup as it could be a FilterNode; for
        // this reason there's this explicit check.
        //
        if (!(delegate instanceof NodePresentationModel) && !(delegate instanceof GenericFilterNode))
          {
            return fallBackName;
          }

        final Displayable displayable = getLookup().lookup(Displayable);
        return (displayable != null) ? displayable.getDisplayName() : fallBackName;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    private Image getIcon (final int sizeType, final @Nonnull Image fallBackImage)
      {
        final Node delegate = getOriginal();
        //
        // If the delegate is not a GenericNode, it is not using an IconProvider for its icon, thus it is directly
        // providing its own icon. It could even have an IconProvider in its Lookup as it could be a FilterNode; for
        // this reason there's this explicit check.
        //
        if (!(delegate instanceof NodePresentationModel) && !(delegate instanceof GenericFilterNode))
          {
            return fallBackImage;
          }

        final IconProvider iconProvider = getLookup().lookup(IconProvider.class);
        final int size = (sizeType == BeanInfo.ICON_COLOR_16x16) ? 16 : 32;
        return (iconProvider != null) ? ImageUtilities.icon2Image(iconProvider.getIcon(size)) : fallBackImage;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @CheckForNull
    public Action getPreferredAction()
      {
        final Node delegate = getOriginal();
        //
        // If the delegate is not a GenericNode, it is not using an ActionProvider, thus it is directly
        // providing its own actions. It could even have an ActionProvider in its Lookup as it could be a FilterNode; for
        // this reason there's this explicit check.
        //
        if (!(delegate instanceof NodePresentationModel) && !(delegate instanceof GenericFilterNode))
          {
            return super.getPreferredAction();
          }

        try
          {
            final ActionProvider actionProvider = getLookup().lookup(ActionProvider);
            return (actionProvider != null) ? actionProvider.getPreferredAction() : super.getPreferredAction();
          }
        catch (NotFoundException e)
          {
            return super.getPreferredAction();
          }
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @CheckForNull
    public Action[] getActions (final boolean context)
      {
        final Node delegate = getOriginal();
        //
        // If the delegate is not a GenericNode, it is not using an ActionProvider, thus it is directly
        // providing its own actions. It could even have an ActionProvider in its Lookup as it could be a FilterNode; for
        // this reason there's this explicit check.
        //
        if (!(delegate instanceof NodePresentationModel) && !(delegate instanceof GenericFilterNode))
          {
            return super.getActions();
          }

        final ActionProvider actionProvider = getLookup().lookup(ActionProvider);
        return (actionProvider != null) ? actionProvider.getActions().toArray(new Action[0]) : super.getActions();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void installIconChangeListener()
      {
        final MutableIconProvider iconProvider = getLookup().lookup(MutableIconProvider.class);
        // FIXME: weak listener? Is the IconProvider allowed to change?

        if (iconProvider != null)
          {
            iconProvider.addPropertyChangeListener(new PropertyChangeListener()
              {
                @Override
                public void propertyChange (final @Nonnull PropertyChangeEvent event)
                  {
                    fireIconChange();
                  }
              });
          }
      }
  }
