/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * Copyright (C) 2006-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://openbluesky.kenai.com
 * SCM: https://bitbucket.org/tidalwave/openbluesky-src
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.nodes;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.beans.BeanInfo;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.awt.Image;
import java.awt.image.BufferedImage;
import javax.swing.Action;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.As;
import it.tidalwave.util.AsException;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.Task;
import it.tidalwave.role.Displayable;
import it.tidalwave.role.IconProvider;
import it.tidalwave.role.MutableDisplayable;
import it.tidalwave.role.Removable;
import it.tidalwave.netbeans.aspect.Aspects;
import it.tidalwave.netbeans.nodes.role.ActionProvider;
import it.tidalwave.role.Composite;
import it.tidalwave.netbeans.nodes.role.HtmlDisplayName;
import it.tidalwave.netbeans.nodes.role.impl.CompositeChildFactory;
import it.tidalwave.role.ui.PresentationModel;
import static it.tidalwave.role.Displayable.Displayable;
import static it.tidalwave.role.IconProvider.IconProvider;
import static it.tidalwave.role.MutableDisplayable.MutableDisplayable;
import static it.tidalwave.role.MutableIconProvider.MutableIconProvider;
import static it.tidalwave.netbeans.nodes.role.ActionProvider.ActionProvider;
import static it.tidalwave.role.Composite.Composite;
import static it.tidalwave.netbeans.nodes.role.Describable.Describable;
import static it.tidalwave.netbeans.nodes.role.HtmlDisplayName.HtmlDisplayName;

/***********************************************************************************************************************
 *
 * An implementation of {@link Node} that provides support for some basic roles and for {@link As}-capable objects.
 * 
 * Roles are searched for in this order:
 * 
 * <ol>
 * <li>in the referred object if it implements {@link As};</li>
 * <li>in the {@code Lookup} of the {@code Node} itself;</li>
 * <li>in the referred object if it implements a role interface</li>
 * </ol>
 * 
 * The supported roles are:
 * 
 * <ol>
 * <li>{@link Displayable} to provide the node display name;</li>
 * <li>{@link MutableDisplayable} to provide the node display name that can be changed at any time;</li>
 * <li>{@link IconProvider} to provide the icon;</li>
 * <li>{@link Removable} whose method {@code remove()} is called upon {@code destroy()}.
 * <li>{@link Composite} to provide the list of children;</li>
 * <li>{@link ActionProvider} to provide contextual menu actions;</li>
 * </ol>
 * 
 * In order to enable the HTML renderer of display names, an instance of {@link HtmlDisplayName} must be present in the
 * node {@code Lookup}.
 * 
 * All the standard actions that can be performed against a node (such as renaming, copying, destroying, etc...) are 
 * appropriately propagated to the proper roles.
 * 
 * Note: at the moment, the copy operation is not supported (there's no role to delegate to).
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@edu.umd.cs.findbugs.annotations.SuppressWarnings("HE_INHERITS_EQUALS_USE_HASHCODE")
public class NodePresentationModel extends AbstractNode implements PresentationModel
  {
    private static final String CLASS = NodePresentationModel.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    private final static BufferedImage EMPTY_ICON = new BufferedImage(16, 16, BufferedImage.TYPE_4BYTE_ABGR);
    
    @Nonnull
    private final Object object;

    private String name;

    /*******************************************************************************************************************
     *
     * Creates a {@code GenericNode} given the referred object. 
     * 
     * @param  object       the referring object
     *
     ******************************************************************************************************************/
    public NodePresentationModel (final @Nonnull Object object)
      {
        this(object, createChildrenFromComposite(object), createObjectLookup(object));  
      }

    /*******************************************************************************************************************
     *
     * Creates an {@code GenericNode} given the referred object, the children and a {@link Lookup}.
     * 
     * @param  object       the referring object
     * @param  children     the children
     * @param  lookup       the {@code Lookup}
     *
     ******************************************************************************************************************/
    public NodePresentationModel (final @Nonnull Object object,
                        final @Nonnull Children children,
                        final @Nonnull Lookup lookup)
      {
        super(children, createLookup(object, lookup));
        this.object = object;

        try
          {
            name = find(object, Displayable).getDisplayName();
          }
        catch (NotFoundException e)
          {
            name = "???"; // do this first
            logger.severe("No Displayable available for %s in %s", this, object);
          }

        installDisplayNameListener();
        installIconChangeListener();
      }
    
    /*******************************************************************************************************************
     *
     * Creates an {@code EhnancedNode} given the referring object, the children and an empty {@link Lookup}.
     * 
     * @param  object       the referring object
     * @param  children     the children
     *
     ******************************************************************************************************************/
    public NodePresentationModel (final @Nonnull Object object, final @Nonnull Children children)
      {
        this(object, children, Lookup.EMPTY);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final void setName (final @Nonnull String name)
      {
        this.name = name;
        super.setName(name);

        Aspects.run(this, new Task<Void, RuntimeException>("setName()")
          {
            @Override
            public Void run()
              {
                try
                  {
                    find(object, MutableDisplayable).setDisplayName(name);
                  }
                catch (NotFoundException e)
                  {
                    // FIXME: do nothing or throw runtime exception?
                  }
                return null;
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public final String getName()
      {
        return name;
      }

//    /*******************************************************************************************************************
//     *
//     * {@inheritDoc}
//     *
//     ******************************************************************************************************************/
//    @Override
//    public final void setDisplayName (final @Nonnull String displayName)
//      {
//        super.setDisplayName(displayName);
//        _setName(displayName);
//      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public final String getDisplayName()
      {
        return name;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String getHtmlDisplayName() 
      {
        return getLookup().lookup(HtmlDisplayName) != null ? getDisplayName() : null;
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String getShortDescription() 
      {
        try 
          {
            return find(object, Describable).getDescription();
          }
        catch (NotFoundException e) 
          {
            return super.getShortDescription();
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Image getIcon (final int sizeType)
      {
        return findIcon((sizeType == BeanInfo.ICON_COLOR_16x16) ? 16 : 32);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Image getOpenedIcon (final int sizeType)
      {
        return findIcon((sizeType == BeanInfo.ICON_COLOR_16x16) ? 16 : 32);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @CheckForNull
    public Action getPreferredAction()
      {
        try
          {
            return find(object, ActionProvider).getPreferredAction();
          }
        catch (NotFoundException e)
          {
            return null;
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @CheckForNull
    public Action[] getActions (final boolean context)
      {
        try
          {
            return find(object, ActionProvider).getActions().toArray(new Action[0]);
          }
        catch (NotFoundException e)
          {
            return new Action[0];
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void destroy()
      throws IOException
      {
        try 
          {
            final Removable removable = find(object, Removable.class);
            
            try
              {
                Aspects.run(this, new Task<Void, Exception>("destroy()")
                  {
                    @Override
                    public Void run()
                      throws Exception
                      {
                        removable.remove();
                        return null;
                      }
                  });
              }
            catch (IOException e)
              {
                throw e;
              }
            catch (Exception e)
              {
                throw new IOException(e);
              }
          }
        catch (NotFoundException e) 
          {
            // ok
          }

        super.destroy();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canCopy()
      {
        return false;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canCut()
      {
        return canCopy() && canDestroy();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canDestroy()
      {
        try
          {
            find(object, Removable.class);
            return true;
          }
        catch (NotFoundException e)
          {
            return false;
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final boolean canRename()
      {
        try
          {
            find(object, MutableDisplayable);
            return true;
          }
        catch (NotFoundException e)
          {
            return false;
          }
      }

   /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public final <T> T as (final @Nonnull Class<T> type)
      {
        return as(type, As.Defaults.throwAsException(type));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> T as (final @Nonnull Class<T> type, final @Nonnull NotFoundBehaviour<T> notFoundBehaviour)
      {
        final T result = getLookup().lookup(type);

        return (result != null) ? result : notFoundBehaviour.run(new NotFoundException(type.getName()));
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String toString()
      {
        return String.format("%s@%x[name=%s, displayName=%s]",
                             getClass().getSimpleName(), System.identityHashCode(this), getName(), name);
      }

    /*******************************************************************************************************************
     *
     * Finds a role.
     * 
     * @param  role                 the role
     * @return                      the role
     * @throws NotFoundException    if no role is found
     *
     ******************************************************************************************************************/
    @Nonnull
    private static <T> T find (final @Nonnull Object object, final @Nonnull Class<T> role)
      throws NotFoundException
      {
        if (object instanceof As)
          {
            try
              {
                return ((As)object).as(role);
              }
            catch (AsException e)
              {
                // ok
              }
          }

        if (object instanceof Lookup.Provider)
          {
            final T lookedUp = ((Lookup.Provider)object).getLookup().lookup(role);

            if (lookedUp != null)
              {
                return lookedUp;
              }
          }

        // Check this after as(...) and lookup()
        if (role.isAssignableFrom(object.getClass()))
          {
            return (T)object;
          }

        throw new NotFoundException();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private Image findIcon (final int size)
      {
        try
          {
            return ImageUtilities.icon2Image(find(object, IconProvider).getIcon(size));
          }
        catch (NotFoundException e)
          {
            return EMPTY_ICON;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void installDisplayNameListener()
      {
        MutableDisplayable displayable = null;
        try
          {
            displayable = find(object, MutableDisplayable);
            // FIXME: weak listener? Is it possible to replace the MutableDisplayable in the Lookup?
            displayable.addPropertyChangeListener(new PropertyChangeListener()
              {
                @Override
                public void propertyChange (final @Nonnull PropertyChangeEvent event)
                  {
                    name = (String)event.getNewValue();
                    fireDisplayNameChange((String)event.getOldValue(), name);
                  }
              });
          }
        // FIXME: Currently ElmoSkosConcept doesn't implement properly MutableDisplayable
        // Keeping this patch until it's fixed
        catch (AbstractMethodError e) 
          {
            final StackTraceElement[] stackTrace = e.getStackTrace();

            if ((stackTrace != null) && (stackTrace.length > 0)) // FIXME
              {
                logger.warning("BUG (harmless) for %s: %s at %s", displayable, e, stackTrace[0]);
              }
            else
              {
                logger.warning("BUG (harmless) for %s: %s", displayable, e);
              }
          }
        // END FIXME
        catch (NotFoundException e)
          {
            // ok, a MutableDisplayable is not mandatory
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void installIconChangeListener()
      {
        try 
          {
            // FIXME: weak listener? Is the IconProvider allowed to change?
            find(object, MutableIconProvider).addPropertyChangeListener(new PropertyChangeListener()
              {
                @Override
                public void propertyChange (final @Nonnull PropertyChangeEvent event)
                  {
                    fireIconChange();
                  }
              });
          }
        catch (NotFoundException e)
          {
            // ok, no MutableIconProvider
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static Lookup createObjectLookup (final @Nonnull Object object)
      {  
        if (object instanceof Lookup.Provider)
          {
            return new ProxyLookup(Lookups.singleton(object), ((Lookup.Provider)object).getLookup());  
          }
        
        return Lookups.singleton(object);
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static Children createChildrenFromComposite (final @Nonnull Object object)
      {
        try 
          {
            return Children.create(new CompositeChildFactory(find(object, Composite)), true);
          }
        catch (NotFoundException e) 
          {
            return Children.LEAF;
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static Lookup createLookup (final @Nonnull Object object, final @Nonnull Lookup lookup)
      {
        return (object instanceof Lookup.Provider) ? new ProxyLookup(((Lookup.Provider)object).getLookup(), lookup) : lookup;
      }

    @Override
    public void addPropertyChangeListener(String string, PropertyChangeListener pl)
      {
        throw new UnsupportedOperationException("Not supported yet.");
      }

    @Override
    public void removePropertyChangeListener(String string, PropertyChangeListener pl) 
      {
        throw new UnsupportedOperationException("Not supported yet.");
      }

    @Override
    public boolean hasListeners(String string)  
      {
        throw new UnsupportedOperationException("Not supported yet.");
      }

    @Override
    public PropertyChangeListener[] getPropertyChangeListeners() 
      {
        throw new UnsupportedOperationException("Not supported yet.");
      }

    @Override
    public PropertyChangeListener[] getPropertyChangeListeners(String string) 
      {
        throw new UnsupportedOperationException("Not supported yet.");
      }
  }
